/*
 *  oscil.h
 *	Copyright Risto Holopainen 2018
 *
 *  Simple wavetable oscillator with linear interpolation.
 *	Expects externally initialized wavetables with guard points.
 *
 	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef OSCIL_H
#define OSCIL_H

#include <cmath>


class oscil
{
	public:
		oscil(int sr, const double *table, int sz, double phase=0);
		~oscil();
		double gen(float frq);
		void gen(double *x, int N, float frq);		// buffer version
		// complex <double> quadrature(float frq);	// TODO
	
	private:
		int sr;
		double phase;	// [0,1)
		const double *wavetable;
		int tab_sz;
		double adr;
		int I;		// integer-address
		double fr;	// fractional-address
		double x;
};



oscil::oscil(int fs, const double *tab, int sz, double ph)
{
	// tab[sz] should contain a guard point (i.e. tab[0] == tab[sz-1]).
	sr = fs;
	tab_sz = sz-1;	
	wavetable = tab;
	phase = ph;
}



oscil::~oscil()
{
	wavetable = 0;
	delete wavetable;
}



double oscil::gen(float frq)
{
// look up in the table and interpolate

	adr = phase*tab_sz;
	I = (int) adr;
	fr = adr - I;
	x = (1-fr) * wavetable[I] + fr*wavetable[I+1];
	phase += frq/sr;
	
	if(phase >= 1.0 || phase < 0.0)
		phase -= floor(phase);	

	return x;
}


void oscil::gen(double *s, int N, float frq)
{
	double omega = frq/sr;
	for(int j=0; j<N; j++)
		{
		adr = phase*tab_sz;
		I = (int) adr;
		fr = adr - I;
		x = fr*wavetable[I] + (1-fr)*wavetable[I+1];	
		s[j] = x;
		phase += omega;	

		if(phase >= 1.0 || phase < 0.0)
			phase -= floor(phase);				
		}
}


#endif
