
/*
 *  waveforms.h
 *	Copyright Risto Holopainen 2018.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 *  For [partial] documentation, see https://ristoid.net/prog.html 
 *	
 *	First argument is a pointer to a wavetable to be created, second arg its length. 
 *
 *	NB: Uses guard points; the length of the wavetable should be odd. First entry = last entry.
 *
 *  void mix(REAL *wt, int sz, const REAL *x, const REAL *y, REAL b=0.5)
 *	waveform::make_sine(REAL *x, int N, int partial=1)
 *	waveform::make_sawtooth(REAL *wt, int sz, int p=15, bool norm=true)
 *	waveform::make_pink(REAL *wt, int sz, int p=15, bool norm=true)
 *	void make_prime(REAL *wt, int sz, int p=10, bool norm=true)
 *	waveform::make_pulse(REAL *x, int N, float alpha=-0.5, int P=15, bool normalize=true)
 *	waveform::make_triangle(REAL *w, int N, int Partials, bool norm=true)
 *	void make_sparse(REAL *wt, int N, int P=55, bool one_over_k=true)
 *	waveform::make_order3(REAL *wt, int sz, int Npartials=15) : partial_k ~ 1/k^3
 *	waveform::make_AM(REAL *w, int N, float a=0) -- a : [0,1] -> sine--square
 * 	waveform::make_bump(REAL *w, int N) -- positive wave
 *  waveform::make_symbump(REAL *w, int N) -- bipolar wave
 *	void make_diffbump(REAL *w, int N, bool norm=true)
 *	void make_halfsine(REAL *wt, int N, int P=25, bool norm=true)
 *	void make_formant(REAL *wt, int N, int cf, bool norm=true)
 *	void make_twinpeaks(REAL *wt, int N, bool norm=true)
 *	void make_band(REAL *wt, uint N, uint lo, uint hi, bool norm=true)
 *	void make_comb(REAL *wt, int N, int offset=1, int P=31, bool norm=true)
 *	void make_noise(REAL *wt, int N, double seed=1/7.)
 *	void make_volterra(REAL *w, int N, bool norm=true)
 *	void make_darboux(REAL *wt, int N, bool norm=true)
 *	void make_octaves(REAL *wt, int N, bool norm=true)
 *	void make_expogliss(REAL *wt, int N, uint p=5, REAL r=8.0, bool norm=true)
 *	
 *	REAL: expects float / double / long double
 */

#ifndef WAVEFORMS_H
#define WAVEFORMS_H


#include <cmath>
#include <cstdio>

namespace waveform
{



const bool NORMALIZE = true;
const bool SKIP_NORM = false;


template <typename REAL>
void mix(REAL *wt, int sz, const REAL *x, const REAL *y, REAL b=0.5)
{
	// mix x[sz] and y[sz] into wt[sz] with balance b
	// such that wt = b*x + (1-b)*y
	
	if(b < 0 || b > 1)
		b = 0.5;
	
	for(int i=0; i<sz; i++)
		{
		wt[i] = b*x[i] + (1-b)*y[i];
		}
}


template <typename REAL>
void make_sine(REAL *wt, int sz, int partial=1)
{
	const int N = sz-1;
	const REAL W = 2*M_PI/N;
	
	for(int i=0; i<sz; i++)
		{
		wt[i] = sin(partial*i*W);
		}
}





template <typename REAL>
void make_sawtooth(REAL *wt, int sz, int p=15, bool norm=true)
{
	const int N = sz-1;
	const REAL W = 2*M_PI/N;
	REAL x, max = 0;
	for(int i=0; i<sz; i++)
		{
		x = 0;
		for(int j=1; j<=p; j++)
			{
			x += sin(j*W*i)/j;
			}
		if(fabs(x)>max)
			max = fabs(x);
		wt[i] = x;
		}
	
	if(norm)
		{
		REAL g = 1.0 / max;
		for(int i=0; i<sz; i++)
			wt[i] *= g;
		}
	wt[N] = wt[0];
}


template <typename REAL>
void make_pink(REAL *wt, int sz, int p=15, bool norm=true)
{
	// all partials, amp = 1/sqrt(k) 

	REAL max = 0, x;
	const int N = sz-1;
	const REAL W = 2*M_PI/N;
	
	for(int i=0; i<N; i++)
		{
		x = 0;
		for(int k=1, j=-1; k<=p; k++)
			{
			x += j*sin(k*W*i) / sqrt(k);
			j *= -1;
			}
		wt[i] = x;
		if(fabs(x) > max) max = fabs(x);
		}
	wt[N] = wt[0];

	if(norm)
		{
		REAL g = 1.0 / max;
		for(int i=0; i<sz; i++)
			wt[i] *= g;
		}
}





template <typename REAL>
void make_prime(REAL *wt, int sz, int p=10, bool norm=true)
{
	// prime number partials
	REAL max = 0, x;
	const int N = sz-1;
	const REAL W = 2*M_PI/N;
	const int max_primes = 18;
	const int primes[max_primes] 
		= {2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61};
	if(p>max_primes) p=max_primes;
	
	for(int i=0; i<N; i++)
		{
		x = 0;
		for(int k=0, j=1; k<p; k++)
			{
			x += j*sin(primes[k]*W*i) / sqrt(k+1);
			j *= -1;
			}
		wt[i] = x;
		if(fabs(x) > max) max = fabs(x);
		}
	wt[N] = wt[0];

	if(norm)
		{
		REAL g = 1.0 / max;
		for(int i=0; i<sz; i++)
			wt[i] *= g;
		}
}



template <typename REAL>
void make_pulse(REAL *x, int sz, float alpha=-0.5, int P=15, bool normalize=true)
{
	// P partials, x[N] = pulsewave, alpha: break point (-1, 1)
	// sz odd: including guard point

	const float a = (alpha > -1 && alpha < 1) ? M_PI*alpha : 0;
	const int N = sz-1;
	REAL S, C, t;
	const REAL invpi = 1.0/M_PI;
	REAL max = 0;
	
	for(int i=0; i<N; i++)
		{
		x[i] = 0;
		t = 2*M_PI*i/N;
		for(int k=1, j=-1; k<P; k++)
			{
			S = -2*sin(k*a);
			C = 2*(cos(k*a) - j);
			j *= -1;
			x[i] += invpi*(S*cos(k*t) + C*sin(k*t))/k;
			}
		x[i] -= a/4;	// k==0
		if(fabs(x[i]) > max)
			max = fabs(x[i]);
		}
	x[N] = x[0];
	if(normalize)
		{
		for(int i=0; i<sz; i++)
			x[i] /= max;
		}
}


template <typename REAL>
void make_triangle(REAL *wt, int tabsz, int Npartials=15, bool norm=true)
{
// x[tabsz] : sum Npartials.  tabsz odd guard point included

	const int N = tabsz -1;
	const int Ps = Npartials;
	REAL max = 0;
	REAL omega = 2*M_PI/N;

	for(int i=0; i<N; i++) 
		{
		wt[i] = 0;
		for(int j=1, k=1; j<=2*Ps; j += 2)	// j: odd partials
			{
			wt[i] += k * sin(j*omega*i) / (j*j);
			k *= -1;
			}
		if(fabs(wt[i]) > max)
			max = fabs(wt[i]);
		}
	wt[N] = wt[0];

	if(norm)
		{
		max += 1E-04;
		for(int i=0; i<tabsz; i++)
			wt[i] /= max;
		}
}



template <typename REAL>
void make_sparse(REAL *wt, int N, int P=55, bool one_over_k=true)
{
	// partials 1,3,6,10,15,21...
	REAL x;
	int M = N-1;
	int maxP = 24;
	int ps[maxP];

	for(int i=1, k=1; i<=maxP; i++)
		{
		ps[i-1] = k;
		k += i+1;
		printf("%2d \t %2d\n", i, ps[i-1]);
		if(k > N/4 || k > P)
			{
			maxP = i;
			break;	//	reached upper frequency limit
			}
		}

	REAL A[maxP];
	if(one_over_k)
		for(int k=0; k<maxP; k++)
			{
			A[k] = 1.0 / (k+1);
			}
	else
		for(int k=0; k<maxP; k++)
			{
			A[k] = 1.0 / ps[k];
			}
	
	REAL max=0;
	const REAL W = 2*M_PI/M;
	for(int i=0; i<M; i++)
		{
		x = 0;
		for(int k=0; k<maxP; k++)
			{
			x += A[k]*sin(W*ps[k]*i);
			}
		wt[i] = x;
		if(fabs(x) > max) max = fabs(x);
		}
	wt[M] = wt[0];
	
	double g = 1.0/max;
	for(int i=0; i<N; i++)
		wt[i] *= g;	
}


template <typename REAL>
void make_order3(REAL *wt, int sz, int Npartials=15)
{
	const int N = sz - 1;
	REAL omega = 2*M_PI/N;
	REAL max = 0;
	
	for(int i=0; i<sz; i++)
		{
		wt[i] = 0;
		for(int k=1; k<=Npartials; k++)
			{
			wt[i] += sin(k*omega*i) / (k*k*k);
			}
		if(fabs(wt[i]) > max)
			max = fabs(wt[i]);
		}
	// normalize
	max += 1E-04;
	for(int i=0; i<sz; i++)
		wt[i] /= max;
}



template <typename REAL>
void make_AM(REAL *wt, int sz, REAL alpha=0)
{
	// sz odd: including guard point
	// alpha: 0 -> hanning, 1 -> square-like

	const int N = sz-1;

	REAL a = (alpha >= 0 && alpha <= 1) ? alpha : 0.0;
	REAL x, u;
	REAL omega = 2*M_PI/N;
	
	for(int i=0; i<N; i++)
		{
		x = cos(i*omega);
		u = (1-a)*x + a*tanh(16*a*x);
		wt[i] = .5*(1 - u);
		}
	wt[N] = wt[0];
}


template <typename REAL>
void make_bump(REAL *w, int N) //-- positive
{
	// smooth function -- has continuous derivatives of all orders.
	int M = N/2;
	REAL x, u, v;
	
	for(int i=0, j; i<M; i++)
		{
		u = i*2.0/N - 1.0;
		v = 1.0 / (1.0 - u*u);
		w[i] = exp(1 - v);
		j = N-1-i;
		w[j] = w[i];
		}
	w[M] = 1.0;
}


template <typename REAL>
void make_symbump(REAL *w, int N) //-- bipolar
{
	// bumpy sinusoid of sorts
	int M = N/2;	
	REAL u, v;
	
	for(int i=0, j; i<M; i++)
		{
		v = i*2.0/M - 1.0;	 // -1..+1
		u = 1.0 / (1.0 - v*v);	
		w[i] = exp(1 - u);
		j = N-1-i;	// reverse
		w[j] = - w[i];
		}
	w[M] = 0.0;
}


template <typename REAL>
void make_diffbump(REAL *w, int N, bool norm=true)
{
// differentiated bump, similar to sawtooth but smoother. bipolar

	int M = N-1;
	double t, g, u, max=0;
	
	for(int i=0; i<M; i++)
		{
		t = (double) 2.0*i/N - 1.0;
		u = (1.0 - t*t);
		g = -2.0*t / (u*u);
		w[i] = g * exp(-1.0/u);
		if(fabs(w[i]) > max) max = fabs(w[i]);
		}
	w[M] = w[0];
	
	if(norm)
		{
		double g = 1.0 / max;
		for(int i=0; i<N; i++)
			w[i] *= g;
		}
}



template <typename REAL>
void make_halfsine(REAL *wt, int N, int P=25, bool norm=true)
{

	const int M = N-1;
	const double W = 2.0*M_PI/M;
	double x, t, a, max=0;
	
	for(int i=0; i<M; i++)
		{
		t = W*i;
		x = 0.5*M_PI*sin(2.0*t);
		for(int k=1; k<P; k += 2)
			{
			a = -4.0/(k*k - 4);
			x += a * cos(k*t);
			}
		wt[i] = x;
		if(fabs(x) > max) max = fabs(x);
		}
	wt[M] = wt[0];
	
	if(norm)
		{
		double g = 1.0 / max;
		for(int i=0; i<N; i++)
			wt[i] *= g;
		}
}


template <typename REAL>
void make_formant(REAL *wt, int N, int cf, bool norm=true)
{
// put a "formant" frequency at partial cf
	int q = 3 + 64/cf;
	const int P = 2*cf + q;
	const int M = N-1;
	const REAL W = 2.0*M_PI/M;
	REAL x, a, df, t, max=0;
	
	for(int i=0; i<M; i++)
		{
		x = 0;
		t = W*i;
		for(int k=1; k<=P; k++)
			{
			df = fabs(k + .5 - cf) * fabs(k - 0.5 - cf);
			a = 1.0 / df;
			x += a * sin(k*t);
			}
		wt[i] = x;
		if(fabs(x) > max) max = fabs(x);
		}
	wt[M] = wt[0];
	
	if(norm)
		{
		REAL g = 1.0 / max;
		for(int i=0; i<N; i++)
			wt[i] *= g;
		}
}



template <typename REAL>
void make_twinpeaks(REAL *wt, int N, bool naive=false, bool norm=true)
{
	// strong 1st, 2nd partials, weak partials >= 3
	const int M = N-1;
	REAL t, u, x, y, s, max=0, mean=0, cumsum=0;
	const REAL c = 2/M_PI;
	const REAL a = c-1;
	const REAL b = 1-2*c;
	REAL p[M];
	const REAL f1 = 2.5*M_PI;
	const REAL f2 = 3.5*M_PI;

	if(naive) 	// p: linear ramp for naive version
		for(int i=0; i<M; i++)
			{
			t = (REAL) i/M;
			p[i] = 1.0 - t;
			}
	else		// p: PARABOLIC, smooth derivative
		for(int i=0; i<M; i++)
			{
			t = (REAL) i/M;
			p[i] = a*t*t + b*t + c;
			}
	
	for(int i=0; i<M; i++)
		{
		t = (REAL) i/M;
		x = sin(f1*t);
		y = sin(f2*t);
		s = p[i]*(x - y);
		wt[i] = s;
		cumsum += s;
		if(fabs(s) > fabs(max)) max = s;
		}

	wt[M] = wt[0];
	cumsum += wt[0];
	mean = cumsum / N;
	
	if(norm)	// also removes DC
		{
		REAL g = 1.0 / (max-mean);
		for(int i=0; i<N; i++)
			wt[i] = (wt[i] - mean) * g;
		}
}



template <typename REAL>
void make_band(REAL *wt, uint N, uint lo, uint hi, bool norm=true)
{
	// partials from lo to hi, amplitudes ~ 1/k
	REAL x, a, max = 0;
	const uint M = N-1;
	const REAL W = 2.0*M_PI/M;
	
	if(hi < lo)	// error !
		return;

	for(uint i=0; i<M; i++)
		{
		x = 0;
		for(uint k=lo; k<hi; k++)
			{
			a = 1.0/k;
			x += a*sin(k*W*i);
			if(fabs(x) > max) max = fabs(x);
			}
		wt[i] = x;
		}
	wt[M] = wt[0];
	
	if(norm)
		{
		REAL g = 1.0 / max;
		for(uint i=0; i<N; i++)
			wt[i] *= g;
		}
}




template <typename REAL>
void make_comb(REAL *wt, int N, int offset=1, int P=31, REAL lambda=0.2, bool norm=true)
{
	// keep every 3d partial starting from offset (suggested values: -2,1,1,2, not 3n) to P
	const int M = N-1;
	REAL x, max=0;
	const REAL W = 2*M_PI/M;
	if(lambda < 0) 	// spectral slope: higher --> steeper, 0=flat
		lambda = 0;

	REAL A[P];
	for(int n=0, k=offset; n<P; n++, k += 3)
		A[n] = exp( -lambda*k );

	for(int i=0; i<M; i++)
		{
		x = 0;
		for(int n=0, k=offset; k<P; n++, k += 3)
			{
			x += A[n]*sin(W*k*i);
			if(fabs(x) > max) max = fabs(x);
			}
		wt[i] = x;			
		}
	wt[M] = wt[0];
		
	if(norm)
		{
		REAL g = 1.0 / max;
		for(int i=0; i<N; i++)
		wt[i] *= g;
		}
}


template <typename REAL>
void make_noise(REAL *wt, int N, double seed=1/7.)
{
	// actually, make logistic equation
	double x = (fabs(seed) < 1) ? seed : 1/7.;
	
	for(int i=0; i<N; i++)
		{
		x = 2*x*x-1;
		wt[i] = x;
		}
}


template <typename REAL>
void make_volterra(REAL *w, int N, bool norm=true)
{
	double t, max=0;
	int M = N-1;
	for(int i=0; i<M; i++)
		{
		t = (double) (i-M/2) / (M/2);
		w[i] = (t == 0) ? 0 : t*t * sin(M_PI / t);
		if( fabs(w[i]) > max )
			max = fabs(w[i]);
		}
	w[M] = w[0];
	if(norm)
		{
		double g = 1.0 / max;
		for(int i=0; i<N; i++)
			w[i] *= g;
		}
}



template <typename REAL>
void make_darboux(REAL *wt, int N, bool norm=true)
{
	// partials such that at least 4 points of fastest period fits in w
	const int M = N-1;	
	const unsigned int P = M/4;

	double W = 2.0*M_PI/M;
	double x, max=0;
	
	const int pmax = 10;
	unsigned int Fact[pmax];
	double fd[pmax];
	Fact[0] = 1;
	fd[0] = 1.0;
	for(int i=1; i<pmax; i++)
		{
		Fact[i] = i*Fact[i-1];
		fd[i] = 1.0 / Fact[i];
		}
	
	for(int i=0; i<M; i++)
		{
		x = cos(W*i);
		for(int k=1; Fact[k]<P && k<pmax; k++)
			{
			x += fd[k] * cos( W*Fact[k]*i );
			}
		wt[i] = x;
		if(fabs(x) > max) max = fabs(x);
		}
	wt[M] = wt[0];

	if(norm)
		{
		double g = 1.0 / max;
			for(int i=0; i<N; i++)
				wt[i] *= g;
		}

}


template <typename REAL>
void make_octaves(REAL *wt, int N, bool norm=true)
{
	/* 
		pathological: infinitely differentiable, yet not analytic 
		at any point --- IF summed over infinitely many partials.
		ref: https://en.wikipedia.org/wiki/Non-analytic_smooth_function
	*/

	const int M = N-1;
	const int P = int( log2(N) - 1);
	double x, max=0;
	const double W = 2.0*M_PI/M;
	double A[P];
	for(int k=0; k<P; k++)
		A[k] = exp(-sqrt(k));

	for(int i=0; i<M; i++)
		{
		x = 0;
		for(int n=0, k; n<P; n++)
			{
			k = 1 << n;
			x += A[n]*cos(k*W*i);
			}
		wt[i] = x;
		if(fabs(x) > max) max = fabs(x);
		}
	wt[M] = wt[0];
	
	if(norm)
		{
		double g = 1.0 / max;
		for(int i=0; i<N; i++)
			wt[i] *= g;
		}
}


template <typename REAL>
void make_expogliss(REAL *wt, int N, uint p=5, REAL r=8.0, bool norm=true)
{
	// P: number of periods of decaying sinusoid; r: glissando ratio > 1
	if(r <= 1.0)
		{
		printf("waveforms::make_expogliss (r=%.2f) must be > 1. (resetting to 2)\n", r);
		r = 2.0;
		}

	const REAL w0 = 2.0 / (r-1);	// r>1
	const REAL gamma = 2*M_PI*p / (w0+1);
	const REAL a = w0*gamma;
	const REAL b = (w0+2)*gamma*cos(gamma*(w0+1));
	const REAL lambda = log(a/b);
	const int M = N-1;	
	REAL t, theta, x, env;
	REAL min=0, max=0, dc=0;

	for(int i=0; i<M; i++)
		{
		t = (REAL) i/M;
		theta = w0*t + t*t;
		x = sin(gamma*theta);
		env = exp(lambda*t);
		wt[i] = env*x;
		if(wt[i] > max) max = wt[i];
		if(wt[i] < min) min = wt[i];
		dc += wt[i];
		}
	wt[M] = wt[0];
	
	if(norm)
		{
		dc /= M;
		REAL gain = (max > fabs(min)) ? 1.0 / (max-dc) : 1.0 / (min-dc);
		for(int i=0; i<N; i++)
			wt[i] = (wt[i] - dc) * gain;
		}
}




template <typename REAL>
void make_chirp(REAL *wt, int N, REAL cf=5, REAL B=12.5)
{
	REAL x=0, t, u;
	const REAL beta = (B > 1) ? B : 1;
	const REAL delta = (cf > 1) ? 2*cf : 2;
	//const REAL W0 = 2*cf;
	const REAL offset = 1.0 / (1.0 + beta);
	REAL g, w, phi=0;
	double max = 0;
	const int M = N-1;
	
	for(int i=0; i<M; i++)
		{
		u = (double) i/M;
		t = 2.0*u - 1.0;
		// w = W0 - delta*u;
		w = delta * (1-u);
		phi += w/M;
		x = sin(phi*2*M_PI);
		g = 1.0/(1.0 + beta*t*t) - offset;
		wt[i] = g*x;
		if(fabs(wt[i]) > max) max = fabs(wt[i]);
		}

	double gain = 1.0 / max;
	for(int i=0; i<N; i++)
		wt[i] *= gain;

	wt[M] = wt[0];
}


template <typename REAL>
void make_diphone(REAL *wt, int N, uint P=5)
{
	// "partial" P
	const REAL twopi = 2.0*M_PI;
	const REAL f1 = twopi*P;
	const REAL A = 1.0 / P;
	REAL x, t;
	const int M = N-1;
	
	for(int i=0; i<M/2; i++)
		{
		t = 2.0 * (double) i/M - 1.0;
		x = sin(twopi*t);
		wt[i] = x; 
		}
	
	for(int i=M/2; i<M; i++)
		{
		t = 2.0 * (double) i/M - 1.0;
		x = A * sin(f1*t);
		wt[i] = x;
		}
	wt[M] = wt[0];
}


} // end namespace WAVEFORM


#endif







